"""Enums for `/api/topics`."""
from .enhanced_enum import EnhancedEnum


class TopicType(EnhancedEnum):
    """Contains constants related for getting certain type of topic."""
    REGULAR_TOPIC = 'Topic'
    CLUB_USER_TOPIC = 'Topics::ClubUserTopic'
    ENTRY_TOPIC = 'Topics::EntryTopic'
    NEWS_TOPIC = 'Topics::NewsTopic'
    ANIME_TOPIC = 'Topics::EntryTopics::AnimeTopic'
    ARTICLE_TOPIC = 'Topics::EntryTopics::ArticleTopic'
    CHARACTER_TOPIC = 'Topics::EntryTopics::CharacterTopic'
    CLUB_PAGE_TOPIC = 'Topics::EntryTopics::ClubPageTopic'
    CLUB_TOPIC = 'Topics::EntryTopics::ClubTopic'
    COLLECTION_TOPIC = 'Topics::EntryTopics::CollectionTopic'
    CONTEST_TOPIC = 'Topics::EntryTopics::ContestTopic'
    COSPLAY_GALLERY_TOPIC = 'Topics::EntryTopics::CosplayGalleryTopic'
    MANGA_TOPIC = 'Topics::EntryTopics::MangaTopic'
    PERSON_TOPIC = 'Topics::EntryTopics::PersonTopic'
    RANOBE_TOPIC = 'Topics::EntryTopics::RanobeTopic'
    CRITIQUE_TOPIC = 'Topics::EntryTopics::CritiqueTopic'
    REVIEW_TOPIC = 'Topics::EntryTopics::ReviewTopic'
    CONTEST_STATUS_TOPIC = 'Topics::NewsTopics::ContestStatusTopic'


class TopicForumType(EnhancedEnum):
    """Contains constants related for getting certain type of forum."""
    ALL = 'all'
    ANIMANGA = 'animanga'
    SITE = 'site'
    GAMES = 'games'
    VN = 'vn'
    CONTESTS = 'contests'
    OFFTOPIC = 'offtopic'
    CLUBS = 'clubs'
    MY_CLUBS = 'my_clubs'
    CRITIQUES = 'critiques'
    NEWS = 'news'
    COLLECTIONS = 'collections'
    ARTICLES = 'articles'
    COSPLAY = 'cosplay'


class TopicLinkedType(EnhancedEnum):
    """Contains constants related for getting certain linked type of topic."""
    ANIME = 'Anime'
    MANGA = 'Manga'
    RANOBE = 'Ranobe'
    CHARACTER = 'Character'
    PERSON = 'Person'
    CLUB = 'Club'
    CLUB_PAGE = 'ClubPage'
    CRITIQUE = 'Critique'
    REVIEW = 'Review'
    CONTEST = 'Contest'
    COSPLAY_GALERY = 'CosplayGallery'
    COLLECTION = 'Collection'
    ARTICLE = 'Article'



"""Special module, which adds some functionality to enum class."""
from enum import Enum


class EnhancedEnum(Enum):
    """Enhanced enum class.

    Adds support for `__str__` for getting enum value
    when casting to string or using in f-string, etc.
    """

    def __str__(self):
        return self.value



"""Enums for `/api/reviews`."""
from .enhanced_enum import EnhancedEnum


class ReviewOpinion(EnhancedEnum):
    """Contains constants related for review opinion type."""
    POSITIVE = 'positive'
    NEUTRAL = 'neutral'
    NEGATIVE = 'negative'



"""Model for `/api/animes|mangas|ranobe/:id/franchise`."""
from typing import List

from pydantic import BaseModel

from .tree_link import TreeLink
from .tree_node import TreeNode


class FranchiseTree(BaseModel):
    """Represents franchise tree entity."""
    links: List[TreeLink]
    nodes: List[TreeNode]
    current_id: int



"""Model for `/api/bans`."""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from .comment import CommentInfo
from .user import UserInfo


class Ban(BaseModel):
    """Represents ban entity."""
    id: int
    user_id: int
    comment: Optional[CommentInfo]
    moderator_id: int
    reason: str
    created_at: datetime
    duration_minutes: int
    user: UserInfo
    moderator: UserInfo



"""Model for `/api/animes|mangas|ranobe/:id/roles`."""
from typing import List, Optional

from pydantic import BaseModel

from .character import CharacterInfo
from .person import PersonInfo


class Role(BaseModel):
    """Represents role info entity."""
    roles: List[str]
    roles_russian: List[str]
    character: Optional[CharacterInfo]
    person: Optional[PersonInfo]



"""Dummy config store class."""
from typing import Optional

from .base import ReturnConfig, Store


class NullStore(Store):
    """Dummy store with an empty implementation of an abstract store class.

    This store is used when no store is provided to the client
    """

    async def save_config(self,
                          app_name: str,
                          client_id: str,
                          client_secret: str,
                          redirect_uri: str,
                          scopes: str,
                          access_token: str,
                          refresh_token: Optional[str] = None,
                          token_expire_at: Optional[int] = None,
                          auth_code: Optional[str] = None):
        pass

    async def fetch_by_access_token(
            self, app_name: str, access_token: str) -> Optional[ReturnConfig]:
        pass

    async def fetch_by_auth_code(self, app_name: str,
                                 auth_code: str) -> Optional[ReturnConfig]:
        pass

    async def delete_token(self, app_name: str, access_token: str):
        pass

    async def delete_all_tokens(self, app_name: str):
        pass



"""Represents `/api/v2/abuse_requests` resource."""
from typing import Any, Dict, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RequestType, ResponseCode
from ..exceptions import ShikimoriAPIResponseError
from ..models import AbuseResponse
from ..utils import Utils
from .base_resource import BaseResource


class AbuseRequests(BaseResource):
    """AbuseRequests resource class.

    Used to represent `/api/v2/abuse_requests` resource
    """

    @method_endpoint('/api/v2/abuse_requests/offtopic')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def offtopic(self, comment_id: int):
        """Marks comment as offtopic.

        :param comment_id: ID of comment to mark as offtopic
        :type comment_id: int

        :return: Abuse response info
        :rtype: Optional[AbuseResponse]
        """
        data_dict = Utils.create_data_dict(comment_id=comment_id)

        response = await self._client.request(
            self._client.endpoints.abuse_offtopic,
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=AbuseResponse)

    @method_endpoint('/api/v2/abuse_requests/review')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def review(self,
                     comment_id: Optional[int] = None,
                     topic_id: Optional[int] = None):
        """Converts comment to review.

        :param comment_id: ID of comment for conversion to review
        :type comment_id: Optional[int]

        :param topic_id: ID of comment's topic
        :type topic_id: Optional[int]

        :return: Abuse response status
        :rtype: bool
        """
        data_dict = Utils.create_data_dict(comment_id=comment_id,
                                           topic_id=topic_id)

        response = await self._client.request(
            self._client.endpoints.abuse_review,
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)

    @method_endpoint('/api/v2/abuse_requests/abuse')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def abuse(self,
                    comment_id: Optional[int] = None,
                    topic_id: Optional[int] = None,
                    reason: Optional[str] = None):
        """Creates abuse about violation of site rules.

        :param comment_id: ID of comment to create abuse request
        :type comment_id: Optional[int]

        :param topic_id: ID of comment's topic
        :type topic_id: Optional[int]

        :param reason: Additional info about violation
        :type reason: Optional[str]

        :return: Abuse response status
        :rtype: bool
        """
        data_dict = Utils.create_data_dict(comment_id=comment_id,
                                           topic_id=topic_id,
                                           reason=reason)

        response = await self._client.request(
            self._client.endpoints.abuse_violation,
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)

    @method_endpoint('/api/v2/abuse_requests/spoiler')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def spoiler(self,
                      comment_id: Optional[int] = None,
                      topic_id: Optional[int] = None,
                      reason: Optional[str] = None):
        """Creates abuse about spoiler in content.

        :param comment_id: ID of comment to create abuse request
        :type comment_id: Optional[int]

        :param topic_id: ID of comment's topic
        :type topic_id: Optional[int]

        :param reason: Additional info about spoiler
        :type reason: Optional[str]

        :return: Abuse response status
        :rtype: bool
        """
        data_dict = Utils.create_data_dict(comment_id=comment_id,
                                           topic_id=topic_id,
                                           reason=reason)

        response = await self._client.request(
            self._client.endpoints.abuse_spoiler,
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)



"""Submodel for `stats.py`."""
from typing import List

from pydantic import BaseModel

from .score import Score


class ScoreList(BaseModel):
    """Represents scores collection of anime/manga."""
    anime: List[Score]
    manga: List[Score]



"""Enums for response status codes."""
from .enhanced_enum import EnhancedEnum


class ResponseCode(EnhancedEnum):
    """Contains response status codes."""
    SUCCESS = 200
    NO_CONTENT = 204
    RETRY_LATER = 429



"""Model for `/api/clubs`."""
from typing import List, Optional

from pydantic import BaseModel

from .anime import AnimeInfo
from .character import CharacterInfo
from .club_image import ClubImage
from .logo import Logo
from .manga import MangaInfo
from .user import UserInfo


class ClubInfo(BaseModel):
    """Represents a club info entity."""
    id: int
    name: str
    logo: Logo
    is_censored: bool
    join_policy: str
    comment_policy: str


class Club(ClubInfo):
    """Represents a club entity."""
    description: Optional[str]
    description_html: str
    mangas: List[MangaInfo]
    characters: List[CharacterInfo]
    thread_id: int
    topic_id: int
    user_role: Optional[str]
    style_id: int
    members: List[UserInfo]
    animes: List[AnimeInfo]
    images: List[ClubImage]



"""Submodel for `club.py`."""
from pydantic import BaseModel


class Logo(BaseModel):
    """Represents club logo links entity."""
    original: str
    main: str
    x96: str
    x73: str
    x48: str



"""Model for `/api/users/:id/favorites`."""
from typing import List

from pydantic import BaseModel

from .favourite import Favourite


class Favourites(BaseModel):
    """Represents collection of favourites by category."""
    animes: List[Favourite]
    mangas: List[Favourite]
    ranobe: List[Favourite]
    characters: List[Favourite]
    people: List[Favourite]
    mangakas: List[Favourite]
    seyu: List[Favourite]
    producers: List[Favourite]



"""Submodel for `character.py`."""
from pydantic import BaseModel

from .image import Image


class Seyu(BaseModel):
    """Represents seyu of character entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str



"""Model for `/api/messages` and submodel for `dialog.py`."""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field

from .linked_topic import LinkedTopic
from .user import UserInfo


class MessageInfo(BaseModel):
    """Represents message info entity."""
    id: int
    kind: str
    read: bool
    body: Optional[str]
    html_body: str
    created_at: datetime
    linked_id: int
    linked_type: Optional[str]
    linked: Optional[LinkedTopic]


class Message(MessageInfo):
    """Represents message entity."""
    from_user: UserInfo = Field(alias='from')
    to_user: UserInfo = Field(alias='to')



"""Represents `/api/people` resource."""
from typing import Any, Dict, List, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import PersonSearchKind
from ..exceptions import ShikimoriAPIResponseError
from ..models import PersonInfo, Person
from ..utils import Utils
from .base_resource import BaseResource


class People(BaseResource):
    """People resource class.

    Used to represent `/api/people` resource
    """

    @method_endpoint('/api/people/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, people_id: int):
        """Returns info about a person.

        :param people_id: ID of person to get info
        :type people_id: int

        :return: Info about a person
        :rtype: Optional[Person]
        """
        response = await self._client.request(
            self._client.endpoints.people(people_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Person)

    @method_endpoint('/api/people/search')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def search(self,
                     search: Optional[str] = None,
                     people_kind: Optional[PersonSearchKind] = None):
        """Returns list of found persons.

        :param search: Search query for persons
        :type search: Optional[str]

        :param people_kind: Kind of person for searching
        :type people_kind: Optional[PersonSearchKind]

        :return: List of found persons
        :rtype: List[PersonInfo]
        """
        query_dict = Utils.create_query_dict(search=search, kind=people_kind)

        response = await self._client.request(
            self._client.endpoints.people_search, query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=PersonInfo)



"""JSON based config store class."""
from json import dumps, loads
from os.path import exists
from typing import Optional

from .base import ConfigsDict, Store
from .memory import MemoryStore


class JSONStore(Store):
    """JSON config store class.

    This class is used for storing configs in JSON file
    """

    __slots__ = ('_file_path',)

    def __init__(self, file_path: str = '.shikithon'):
        super().__init__()
        self._file_path = file_path

    async def _read_from_file(self) -> Optional[ConfigsDict]:
        if not exists(self._file_path):
            return None

        with open(self._file_path, 'r', encoding='utf-8') as file:
            return loads(file.read())

    async def _write_to_file(self, configs: ConfigsDict):
        try:
            with open(self._file_path, 'w', encoding='utf-8') as file:
                file.write(dumps(configs, indent=4))
            return True
        except IOError:
            return False

    async def save_config(self,
                          app_name: str,
                          client_id: str,
                          client_secret: str,
                          redirect_uri: str,
                          scopes: str,
                          access_token: str,
                          refresh_token: Optional[str] = None,
                          token_expire_at: Optional[int] = None,
                          auth_code: Optional[str] = None):
        async with MemoryStore(await self._read_from_file()) as ms:
            await ms.save_config(app_name=app_name,
                                 client_id=client_id,
                                 client_secret=client_secret,
                                 redirect_uri=redirect_uri,
                                 scopes=scopes,
                                 access_token=access_token,
                                 refresh_token=refresh_token,
                                 token_expire_at=token_expire_at,
                                 auth_code=auth_code)

            await self._write_to_file(ms.configs)

    async def fetch_by_access_token(self, app_name: str, access_token: str):
        async with MemoryStore(await self._read_from_file()) as ms:
            return await ms.fetch_by_access_token(app_name=app_name,
                                                  access_token=access_token)

    async def fetch_by_auth_code(self, app_name: str, auth_code: str):
        async with MemoryStore(await self._read_from_file()) as ms:
            return await ms.fetch_by_auth_code(app_name=app_name,
                                               auth_code=auth_code)

    async def delete_token(self, app_name: str, access_token: str):
        async with MemoryStore(await self._read_from_file()) as ms:
            await ms.delete_token(app_name=app_name, access_token=access_token)

            await self._write_to_file(ms.configs)

    async def delete_all_tokens(self, app_name: str):
        async with MemoryStore(await self._read_from_file()) as ms:
            await ms.delete_all_tokens(app_name)

            await self._write_to_file(ms.configs)



"""Represents `/api/comments` resource."""
from typing import Any, Dict, List, Optional, cast

from loguru import logger

from ..decorators import exceptions_handler, method_endpoint
from ..enums import CommentableCreateType, CommentableType, RequestType
from ..exceptions import ShikimoriAPIResponseError
from ..models import Comment
from ..utils import Utils
from .base_resource import BaseResource

DICT_NAME = 'comment'


class Comments(BaseResource):
    """Comments resource class.

    Used to represent `/api/comments` resource
    """

    @method_endpoint('/api/comments')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      commentable_id: int,
                      commentable_type: CommentableType,
                      page: Optional[int] = None,
                      limit: Optional[int] = None,
                      desc: Optional[int] = None):
        """Returns list of comments.

        :param commentable_id: ID of entity to get comment
        :type commentable_id: int

        :param commentable_type: Type of entity to get comment
        :type commentable_type: CommentableType

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param desc: Status of description in request. Can be 1 or 0
        :type desc: Optional[int]

        :return: List of comments
        :rtype: List[Comment]
        """
        query_dict = Utils.create_query_dict(page=page,
                                             limit=limit,
                                             commentable_id=commentable_id,
                                             commentable_type=commentable_type,
                                             desc=desc)

        response = await self._client.request(self._client.endpoints.comments,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Comment)

    @method_endpoint('/api/comments/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, comment_id: int):
        """Returns comment info.

        :param comment_id: ID of comment
        :type comment_id: int

        :return: Comment info
        :rtype: Optional[Comment]
        """
        response = await self._client.request(
            self._client.endpoints.comment(comment_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Comment)

    @method_endpoint('/api/comments')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def create(self,
                     body: str,
                     commentable_id: int,
                     commentable_type: CommentableCreateType,
                     is_offtopic: Optional[bool] = None,
                     broadcast: Optional[bool] = None) -> Optional[Comment]:
        """Creates comment.

        When commentable_type set to other than Topic/User,
        comment is attached to commentable main topic

        :param body: Body of comment
        :type body: str

        :param commentable_id: ID of entity to comment on
        :type commentable_id: int

        :param commentable_type: Type of entity to comment on
        :type commentable_type: CommentableCreateType

        :param is_offtopic: Status of offtopic
        :type is_offtopic: Optional[bool]

        :param broadcast: Broadcast comment in club’s topic status
        :type broadcast: Optional[bool]

        :return: Created comment info
        :rtype: Optional[Comment]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           body=body,
                                           commentable_id=commentable_id,
                                           commentable_type=commentable_type,
                                           is_offtopic=is_offtopic)

        if isinstance(broadcast, bool):
            logger.debug('Adding a broadcast value to a data_dict')
            data_dict['broadcast'] = broadcast

        response = await self._client.request(self._client.endpoints.comments,
                                              data=data_dict,
                                              request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Comment)

    @method_endpoint('/api/comments/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def update(self, comment_id: int, body: str):
        """Updates comment.

        To change is_offtopic field,
        use /api/v2/abuse_requests method

        :param comment_id: ID of comment to update
        :type comment_id: int

        :param body: New body of comment
        :type body: str

        :return: Updated comment info
        :rtype: Optional[Comment]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME, body=body)

        response = await self._client.request(
            self._client.endpoints.comment(comment_id),
            data=data_dict,
            request_type=RequestType.PATCH)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Comment)

    @method_endpoint('/api/comments/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete(self, comment_id: int):
        """Deletes comment.

        Instead of returning just response code,
        API method returns dictionary with "notice"
        field, so it's being logged out with INFO level

        :param comment_id: ID of comment to delete
        :type comment_id: int

        :return: Status of comment deletion
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.comment(comment_id),
            request_type=RequestType.DELETE)

        logger.info(response)

        return True



"""Base classes for config store."""
from typing import Dict, List, Optional, TypedDict


class Token(TypedDict, total=False):
    scopes: str
    access_token: str
    refresh_token: Optional[str]
    token_expire_at: Optional[int]
    auth_code: Optional[str]


class Config(TypedDict, total=False):
    client_id: str
    client_secret: str
    redirect_uri: str
    tokens: List[Token]


class ReturnConfig(Config):
    app_name: str


ConfigsDict = Dict[str, Config]


class Store:
    """Abstract class for config store.

    This class is used to create custom stores by overriding abstract methods
    """

    __slots__ = ('_closed',)

    def __init__(self):
        self._closed = True

    @property
    def closed(self):
        """Returns close status of store.

        :return: True if store is closed, False otherwise
        :rtype: bool
        """
        return self._closed

    async def save_config(self,
                          app_name: str,
                          client_id: str,
                          client_secret: str,
                          redirect_uri: str,
                          scopes: str,
                          access_token: str,
                          refresh_token: Optional[str] = None,
                          token_expire_at: Optional[int] = None,
                          auth_code: Optional[str] = None):
        """Saves config using the method of the selected store class.

        :param app_name: Application name
        :type app_name: str

        :param client_id: Client ID
        :type client_id: str

        :param client_secret: Client secret
        :type client_secret: str

        :param redirect_uri: Redirect URI
        :type redirect_uri: str

        :param scopes: Application scopes
        :type scopes: str

        :param access_token: Access token
        :type access_token: str

        :param refresh_token: Refresh token
        :type refresh_token: Optional[str]

        :param token_expire_at: Token expire time
        :type token_expire_at: Optional[int]

        :param auth_code: Auth code
        :type auth_code: Optional[str]
        """
        raise NotImplementedError

    async def fetch_by_access_token(
            self, app_name: str, access_token: str) -> Optional[ReturnConfig]:
        """Fetches config by access token.

        :param app_name: Application name
        :type app_name: str

        :param access_token: Access token
        :type access_token: str

        :return: Config dictionary
        :rtype: Optional[ReturnConfig]
        """
        raise NotImplementedError

    async def fetch_by_auth_code(self, app_name: str,
                                 auth_code: str) -> Optional[ReturnConfig]:
        """Fetches config by auth code.

        :param app_name: Application name
        :type app_name: str

        :param auth_code: Auth code
        :type auth_code: str

        :return: Config dictionary
        :rtype: Optional[ReturnConfig]
        """
        raise NotImplementedError

    async def delete_token(self, app_name: str, access_token: str):
        """Deletes token from config.

        :param app_name: Application name
        :type app_name: str

        :param access_token: Access token
        :type access_token: str
        """
        raise NotImplementedError

    async def delete_all_tokens(self, app_name: str):
        """Deletes all tokens from config.

        :param app_name: Application name
        :type app_name: str
        """
        raise NotImplementedError

    async def open(self):
        """Opens store and returns self.

        :return: Store instance
        :rtype: Store
        """
        self._closed = False
        return self

    async def close(self):
        """Closes store."""
        self._closed = True

    async def __aenter__(self):
        """Async context manager entry point.

        :return: Store instance
        :rtype: Store
        """
        return await self.open()

    async def __aexit__(self, *args):
        """Async context manager exit point."""
        await self.close()



"""Config store for shikithon library."""

from .base import Store
from .json import JSONStore
from .memory import MemoryStore
from .null import NullStore

__all__ = ['Store', 'NullStore', 'MemoryStore', 'JSONStore']



"""Model for `/api/reviews`."""
from datetime import datetime
from typing import Literal, Optional, Union

from pydantic import BaseModel


class Review(BaseModel):
    """Represents review entity."""
    id: int
    user_id: int
    anime_id: Optional[int]
    manga_id: Optional[int]
    body: str
    opinion: Union[Literal['positive'], Literal['neutral'], Literal['negative']]
    is_written_before_release: bool
    created_at: datetime
    updated_at: datetime
    comments_count: int
    cached_votes_up: int
    cached_votes_down: int
    changed_at: Optional[datetime]



"""Represents `/api/forums` resource."""
from typing import Any, Dict, List, cast

from ..decorators import exceptions_handler, method_endpoint
from ..exceptions import ShikimoriAPIResponseError
from ..models import Forum
from ..utils import Utils
from .base_resource import BaseResource


class Forums(BaseResource):
    """Forums resource class.

    Used to represent `/api/forums` resource
    """

    @method_endpoint('/api/forums')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self):
        """Returns list of forums.

        :returns: List of forums
        :rtype: List[Forum]
        """
        response = await self._client.request(self._client.endpoints.forums)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Forum)



"""Submodel for `user.py`."""
from pydantic import BaseModel


class UserImage(BaseModel):
    """Represents user profile links entity."""
    x160: str
    x148: str
    x80: str
    x64: str
    x48: str
    x32: str
    x16: str



"""Model for `/api/animes/:anime_id/videos`."""
from typing import Optional

from pydantic import BaseModel


class Video(BaseModel):
    """Represents a video entity."""
    id: int
    url: str
    image_url: str
    player_url: str
    name: Optional[str]
    kind: str
    hosting: str



"""Model for `/api/user_images`."""
from pydantic import BaseModel


class CreatedUserImage(BaseModel):
    """Represents created user image entity."""
    id: int
    preview: str
    url: str
    bbcode: str



"""Represents `/api/user_images` resource."""
from typing import Any, Dict, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RequestType
from ..exceptions import ShikimoriAPIResponseError
from ..models import CreatedUserImage
from ..utils import Utils
from .base_resource import BaseResource


class UserImages(BaseResource):
    """UserImages resource class.

    Used to represent `/api/user_images` resource
    """

    @method_endpoint('/api/user_images')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def create(self, image_path: str, linked_type: Optional[str] = None):
        """Creates an user image.

        :param image_path: Path or URL to image to create on server
        :type image_path: str

        :param linked_type: Type of linked image
        :type linked_type: Optional[str]

        :return: Created image info
        :rtype: Optional[CreatedUserImage]
        """
        image_data = await Utils.get_image_data(image_path)
        data_dict = Utils.create_data_dict(image=image_data,
                                           linked_type=linked_type)
        form_data = Utils.create_form_data(data_dict)

        response = await self._client.request(
            self._client.endpoints.user_images,
            form_data=form_data,
            request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=CreatedUserImage)



"""Represents `/api/bans` resource."""
from typing import Any, Dict, List, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..exceptions import ShikimoriAPIResponseError
from ..models import Ban
from ..utils import Utils
from .base_resource import BaseResource


class Bans(BaseResource):
    """Bans resource class.

    Used to represent `/api/bans` resource
    """

    @method_endpoint('/api/bans')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      page: Optional[int] = None,
                      limit: Optional[int] = None):
        """Returns list of recent bans on Shikimori.

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :return: List of recent bans
        :rtype: List[Ban]
        """
        query_dict = Utils.create_query_dict(page=page, limit=limit)

        response = await self._client.request(self._client.endpoints.bans_list,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Ban)



"""Enums for types of requests."""
from .enhanced_enum import EnhancedEnum


class RequestType(EnhancedEnum):
    """Contains types of requests."""
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    DELETE = 'DELETE'



"""Models for the Shikimori API."""

from .abuse_response import AbuseResponse
from .achievement import Achievement
from .anime import AnimeInfo, Anime, CharacterAnime
from .ban import Ban
from .calendar_event import CalendarEvent
from .character import CharacterInfo, Character
from .club import ClubInfo, Club
from .club_image import ClubImage
from .comment import CommentInfo, Comment
from .constants import (AnimeConstants, ClubConstants, MangaConstants,
                        SmileyConstant, UserRateConstants)
from .created_user_image import CreatedUserImage
from .critique import Critique
from .dialog import Dialog
from .favourites import Favourites
from .forum import Forum
from .franchise_tree import FranchiseTree
from .genre import Genre
from .history import History
from .link import Link
from .manga import MangaInfo, Manga, CharacterManga
from .message import MessageInfo, Message
from .person import PersonInfo, Person
from .publisher import Publisher
from .ranobe import RanobeInfo, Ranobe, CharacterRanobe
from .relation import Relation
from .review import Review
from .role import Role
from .screenshot import Screenshot
from .studio import Studio
from .style import Style
from .topic import Topic, TopicUpdate
from .unread_messages import UnreadMessages
from .user import UserInfo, UserBrief, User
from .user_list import UserList
from .user_rate import UserRate
from .video import Video

__all__ = [
    'AbuseResponse', 'AnimeInfo', 'Anime', 'CharacterAnime', 'Achievement',
    'Ban', 'CalendarEvent', 'CharacterInfo', 'Character', 'ClubInfo', 'Club',
    'ClubImage', 'CommentInfo', 'Comment', 'AnimeConstants', 'ClubConstants',
    'MangaConstants', 'SmileyConstant', 'UserRateConstants', 'CreatedUserImage',
    'Critique', 'Dialog', 'Favourites', 'Forum', 'FranchiseTree', 'Genre',
    'History', 'Link', 'MangaInfo', 'Manga', 'CharacterManga', 'MessageInfo',
    'Message', 'PersonInfo', 'Person', 'Publisher', 'RanobeInfo', 'Ranobe',
    'CharacterRanobe', 'Relation', 'Review', 'Role', 'Screenshot', 'Studio',
    'Style', 'Topic', 'TopicUpdate', 'UnreadMessages', 'UserInfo', 'UserBrief',
    'User', 'UserList', 'UserRate', 'Video'
]



"""Enums for `/api/clubs`."""
from .enhanced_enum import EnhancedEnum


class JoinPolicy(EnhancedEnum):
    """Contains constants related for join policy setting."""
    FREE = 'free'
    MEMBER_INVITE = 'member_invite'
    ADMIN_INVITE = 'admin_invite'
    OWNER_INVITE = 'owner_invite'


class CommentPolicy(EnhancedEnum):
    """Contains constants related for comment policy setting."""
    FREE = 'free'
    MEMBERS = 'members'
    ADMINS = 'admins'


class TopicPolicy(EnhancedEnum):
    """Contains constants related for topic policy setting."""
    MEMBERS = 'members'
    ADMINS = 'admins'


class PagePolicy(EnhancedEnum):
    """Contains constants related for page policy setting."""
    MEMBERS = 'members'
    ADMINS = 'admins'


class ImageUploadPolicy(EnhancedEnum):
    """Contains constants related for image upload policy setting."""
    MEMBERS = 'members'
    ADMINS = 'admins'



"""Enums for `/api/user_rates` and `/api/v2/user_rates`."""
from .enhanced_enum import EnhancedEnum


class UserRateType(EnhancedEnum):
    """Contains constants related for getting certain type of user rate."""
    ANIME = 'anime'
    MANGA = 'manga'


class UserRateTarget(EnhancedEnum):
    """Contains constants related for getting
    certain target type of user rate."""
    ANIME = 'Anime'
    MANGA = 'Manga'


class UserRateStatus(EnhancedEnum):
    """Contains constants related for getting
    certain status of item in user rates list.
    """
    PLANNED = 'planned'
    WATCHING = 'watching'
    REWATCHING = 'rewatching'
    COMPLETED = 'completed'
    ON_HOLD = 'on_hold'
    DROPPED = 'dropped'



"""Submodel for `anime.py`."""
from typing import Optional

from pydantic import BaseModel


class Genre(BaseModel):
    """Represents genre of anime entity."""
    id: int
    name: str
    russian: str
    kind: Optional[str]



"""Model for `/api/animes|mangas|ranobe/:id/related`."""
from typing import Optional, Union

from pydantic import BaseModel

from .anime import AnimeInfo
from .manga import MangaInfo
from .ranobe import RanobeInfo


class Relation(BaseModel):
    """Represents relation entity."""
    relation: str
    relation_russian: str
    anime: Optional[AnimeInfo]
    manga: Optional[Union[MangaInfo, RanobeInfo]]



"""Represents `/api/friends` resource."""
from loguru import logger

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RequestType
from ..exceptions import ShikimoriAPIResponseError
from .base_resource import BaseResource


class Friends(BaseResource):
    """Friends resource class.

    Used to represent `/api/friends` resource
    """

    @method_endpoint('/api/friends/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def create(self, user_id: int):
        """Creates (adds) new friend by ID.

        :param user_id: ID of a user to create (add)
        :type user_id: int

        :return: Status of create (addition)
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.friend(user_id),
            request_type=RequestType.POST)

        logger.info(response)

        return True

    @method_endpoint('/api/friends/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def destroy(self, user_id: int):
        """Destroys (removes) current friend by ID.

        :param user_id: ID of a user to destroy (remove)
        :type user_id: int

        :return: Status of destroy (removal)
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.friend(user_id),
            request_type=RequestType.DELETE)

        logger.info(response)

        return True



"""Model for `/api/ranobe`."""
from datetime import date
from typing import List, Optional

from pydantic import BaseModel, validator

from .genre import Genre
from .image import Image
from .publisher import Publisher
from .user_rate import UserRate
from .user_rate_score import UserRateScore
from .user_rate_status import UserRateStatus


class RanobeInfo(BaseModel):
    """Represents ranobe info entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str
    kind: str
    score: float
    status: str
    volumes: int
    chapters: int
    aired_on: Optional[date]
    released_on: Optional[date]

    # pylint: disable=E0213
    @validator('kind')
    def kind_validator(cls, v):
        if 'novel' not in v:
            raise ValueError(f'Invalid kind. Got "{v}"'
                             f' but expected kind, containing "novel"')
        return v


class Ranobe(RanobeInfo):
    """Represents ranobe entity."""
    english: List[Optional[str]]
    japanese: List[Optional[str]]
    synonyms: List[str]
    license_name_ru: Optional[str]
    description: Optional[str]
    description_html: str
    description_source: Optional[str]
    franchise: Optional[str]
    favoured: bool
    anons: bool
    ongoing: bool
    thread_id: Optional[int]
    topic_id: Optional[int]
    myanimelist_id: int
    rates_scores_stats: List[UserRateScore]
    rates_statuses_stats: List[UserRateStatus]
    licensors: List[str]
    genres: List[Genre]
    publishers: List[Publisher]
    user_rate: Optional[UserRate]


class CharacterRanobe(RanobeInfo):
    """Represents a character ranobe info entity."""
    roles: List[str]
    role: str



"""Submodel for `topic.py`."""
from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel

from .anime import AnimeInfo
from .manga import MangaInfo
from .ranobe import RanobeInfo
from .user import UserInfo


class Critique(BaseModel):
    """Represents critique entity.

    Can be found in topics with type
    `Topics::EntryTopics::CritiqueTopic`
    """
    id: int
    target: Optional[Union[AnimeInfo, MangaInfo, RanobeInfo]]
    user: UserInfo
    votes_count: int
    votes_for: int
    body: str
    html_body: str
    overall: Optional[int]
    storyline: Optional[int]
    music: Optional[int]
    characters: Optional[int]
    animation: Optional[int]
    created_at: datetime



"""Represents `/api/calendar` resource."""
from typing import Any, Dict, List, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import AnimeCensorship
from ..exceptions import ShikimoriAPIResponseError
from ..models import CalendarEvent
from ..utils import Utils
from .base_resource import BaseResource


class Calendars(BaseResource):
    """Calendars resource class.

    Used to represent `/api/calendar` resource
    """

    @method_endpoint('/api/calendar')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self, censored: Optional[AnimeCensorship] = None):
        """Returns current calendar events.

        :param censored: Status of censorship for events
        :type censored: Optional[AnimeCensorship]

        :return: List of calendar events
        :rtype: List[CalendarEvent]
        """
        query_dict = Utils.create_query_dict(censored=censored)

        response = await self._client.request(self._client.endpoints.calendar,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=CalendarEvent)



"""Enums for shikithon API class."""

from .anime import (AnimeCensorship, AnimeDuration, AnimeKind, AnimeList,
                    AnimeOrder, AnimeRating, AnimeStatus, AnimeTopicKind)
from .club import (CommentPolicy, ImageUploadPolicy, JoinPolicy, PagePolicy,
                   TopicPolicy)
from .comment import CommentableCreateType, CommentableType
from .favorite import FavoriteLinkedType
from .history import HistoryTargetType
from .manga import (MangaCensorship, MangaKind, MangaList, MangaOrder,
                    MangaStatus)
from .message import MessageType
from .person import PersonKind, PersonSearchKind
from .ranobe import RanobeCensorship, RanobeList, RanobeOrder, RanobeStatus
from .request import RequestType
from .response import ResponseCode
from .review import ReviewOpinion
from .style import StyleOwner
from .topic import TopicForumType, TopicLinkedType, TopicType
from .user_rate import UserRateStatus, UserRateTarget, UserRateType
from .video import VideoKind

__all__ = [
    'AnimeCensorship', 'AnimeDuration', 'AnimeKind', 'AnimeTopicKind',
    'AnimeList', 'AnimeOrder', 'AnimeRating', 'AnimeStatus', 'CommentPolicy',
    'ImageUploadPolicy', 'JoinPolicy', 'PagePolicy', 'TopicPolicy',
    'CommentableType', 'CommentableCreateType', 'HistoryTargetType',
    'FavoriteLinkedType', 'MangaCensorship', 'MangaKind', 'MangaList',
    'MangaOrder', 'MangaStatus', 'MessageType', 'PersonKind',
    'PersonSearchKind', 'RanobeCensorship', 'RanobeList', 'RanobeOrder',
    'RanobeStatus', 'RequestType', 'ResponseCode', 'ReviewOpinion',
    'StyleOwner', 'TopicForumType', 'TopicLinkedType', 'TopicType',
    'UserRateStatus', 'UserRateTarget', 'UserRateType', 'VideoKind'
]



"""Represents `/api/messages` resource."""
from typing import Any, Dict, List, Optional, Union, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import MessageType, RequestType, ResponseCode
from ..exceptions import ShikimoriAPIResponseError
from ..models import Message
from ..utils import Utils
from .base_resource import BaseResource

DICT_NAME = 'message'
PRIVATE_DM = 'Private'


class Messages(BaseResource):
    """Messages resource class.

    Used to represent `/api/messages` resource
    """

    @method_endpoint('/api/messages/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, message_id: int):
        """Returns message info.

        :param message_id: ID of message to get info
        :type message_id: int

        :return: Message info
        :rtype: Optional[Message]
        """
        response = await self._client.request(
            self._client.endpoints.message(message_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Message)

    @method_endpoint('/api/messages')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def create(self, body: str, from_id: int,
                     to_id: int) -> Optional[Message]:
        """Creates message.

        :param body: Body of message (Need to have length >= 2)
        :type body: str

        :param from_id: Sender ID
        :type from_id: int

        :param to_id: Reciver ID
        :type to_id: int

        :return: Created message info
        :rtype: Optional[Message]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           body=body,
                                           from_id=from_id,
                                           kind=PRIVATE_DM,
                                           to_id=to_id)

        response = await self._client.request(self._client.endpoints.messages,
                                              data=data_dict,
                                              request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Message)

    @method_endpoint('/api/messages/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def update(self, message_id: int, body: str):
        """Updates message.

        :param message_id: ID of message to update
        :type message_id: int

        :param body: New body of message
        :type body: str

        :return: Updated message info
        or None if message cannot be updated
        :rtype: Optional[Message]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME, body=body)

        response = await self._client.request(
            self._client.endpoints.message(message_id),
            data=data_dict,
            request_type=RequestType.PATCH)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Message)

    @method_endpoint('/api/messages/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete(self, message_id: int):
        """Deletes message.

        :param message_id: ID of message to delete
        :type message_id: int

        :return: Status of message deletion
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.message(message_id),
            request_type=RequestType.DELETE)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.NO_CONTENT)

    @method_endpoint('/api/messages/mark_read')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def mark_read(self,
                        message_ids: Optional[Union[int, List[int]]] = None,
                        is_read: Optional[bool] = None):
        """Marks read/unread selected messages.

        :param message_ids: ID(s) of messages to mark read/unread
        :type message_ids: Optional[Union[int, List[int]]]

        :param is_read: Status of message (read/unread)
        :type is_read: Optional[bool]

        :return: Status of messages read/unread
        :rtype: bool
        """
        data_dict = Utils.create_query_dict(ids=message_ids, is_read=is_read)

        response = await self._client.request(
            self._client.endpoints.messages_mark_read,
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)

    @method_endpoint('/api/messages/read_all')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def read_all(self, message_type: MessageType):
        """Reads all messages on current user's account.

        This method accepts as message type
        only 'news' and 'notifications'

        :param message_type: Type of messages to read
        :type message_type: MessageType

        :return: Status of messages read
        :rtype: bool
        """
        data_dict = Utils.create_query_dict(type=message_type)

        response = await self._client.request(
            self._client.endpoints.messages_read_all,
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)

    @method_endpoint('/api/messages/delete_all')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete_all(self, message_type: MessageType):
        """Deletes all messages on current user's account.

        This method accepts as message type
        only 'news' and 'notifications'

        :param message_type: Type of messages to delete
        :type message_type: MessageType

        :return: Status of messages deletion
        :rtype: bool
        """
        data_dict = Utils.create_query_dict(type=message_type)

        response = await self._client.request(
            self._client.endpoints.messages_delete_all,
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)



"""Represents `/api/stats` resource."""
from typing import List, cast

from ..decorators import exceptions_handler, method_endpoint
from ..exceptions import ShikimoriAPIResponseError
from .base_resource import BaseResource


class Stats(BaseResource):
    """Stats resource class.

    Used to represent `/api/stats` resource
    """

    @method_endpoint('/api/stats/active_users')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def active_users(self):
        """Returns list of IDs of active users.

        :return: List of IDs of active users
        :rtype: List[int]
        """
        response = await self._client.request(
            self._client.endpoints.active_users)

        return cast(List[int], response)



"""Represents `/api/user_rates` and `/api/v2/user_rates` resources."""
from typing import Any, Dict, List, Optional, cast

from loguru import logger

from ..decorators import exceptions_handler, method_endpoint
from ..enums import (RequestType, ResponseCode, UserRateStatus, UserRateTarget,
                     UserRateType)
from ..exceptions import ShikimoriAPIResponseError
from ..models import UserRate
from ..utils import Utils
from .base_resource import BaseResource

DICT_NAME = 'user_rate'


class UserRates(BaseResource):
    """UserRates resource class.

    Used to represent `/api/user_rates` and `/api/v2/user_rates` resources
    """

    @method_endpoint('/api/v2/user_rates')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      user_id: int,
                      target_id: Optional[int] = None,
                      target_type: Optional[UserRateTarget] = None,
                      status: Optional[UserRateStatus] = None,
                      page: Optional[int] = None,
                      limit: Optional[int] = None):
        """Returns list of user rates.

        When passing target_id, target_type is required

        Also there is a strange API behavior, so when pass nothing,
        endpoint not working.
        However, docs shows that page/limit ignored when user_id is set (bruh)

        :param user_id: ID of user to get rates for
        :type user_id: int

        :param target_id: ID of anime/manga to get rates for
        :type target_id: Optional[int]

        :param target_type: Type of target_id to get rates for
        :type target_type: Optional[UserRateTarget]

        :param status: Status of target_type to get rates for
        :type status: Optional[UserRateStatus]

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
            (This field is ignored when user_id is set)
        :type limit: Optional[int]

        :return: List with info about user rates
        :rtype: List[UserRate]
        """
        query_dict = Utils.create_query_dict(user_id=user_id,
                                             target_id=target_id,
                                             target_type=target_type,
                                             status=status,
                                             page=page,
                                             limit=limit)

        response = await self._client.request(self._client.endpoints.user_rates,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=UserRate)

    @method_endpoint('/api/v2/user_rates/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, rate_id: int):
        """Returns info about user rate.

        :param rate_id: ID of rate to get
        :type rate_id: int

        :return: Info about user rate
        :rtype: Optional[UserRate]
        """
        response = await self._client.request(
            self._client.endpoints.user_rate(rate_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=UserRate)

    @method_endpoint('/api/v2/user_rates')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def create(self,
                     user_id: int,
                     target_id: int,
                     target_type: UserRateTarget,
                     status: Optional[UserRateStatus] = None,
                     score: Optional[int] = None,
                     chapters: Optional[int] = None,
                     episodes: Optional[int] = None,
                     volumes: Optional[int] = None,
                     rewatches: Optional[int] = None,
                     text: Optional[str] = None):
        """Creates new user rate and return info about it.

        :param user_id: ID of user to create user rate for
        :type user_id: int

        :param target_id: ID of target to create user rate for
        :type target_id: int

        :param target_type: Type of target_id to create user rate for
            (Anime or Manga)
        :type target_type: UserRateTarget

        :param status: Status of target
        :type status: Optional[UserRateStatus]

        :param score: Score of target
        :type score: Optional[int]

        :param chapters: Watched/read chapters of target
        :type chapters: Optional[int]

        :param episodes: Watched/read episodes of target
        :type episodes: Optional[int]

        :param volumes: Watched/read volumes of target
        :type volumes: Optional[int]

        :param rewatches: Number of target rewatches
        :type rewatches: Optional[int]

        :param text: Text note for user rate
        :type text: Optional[str]

        :return: Info about new user rate
        :rtype: Optional[UserRate]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           user_id=user_id,
                                           target_id=target_id,
                                           target_type=target_type,
                                           status=status,
                                           score=score,
                                           chapters=chapters,
                                           episodes=episodes,
                                           volumes=volumes,
                                           rewatches=rewatches,
                                           text=text)

        response = await self._client.request(self._client.endpoints.user_rates,
                                              data=data_dict,
                                              request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=UserRate)

    @method_endpoint('/api/v2/user_rates/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def update(self,
                     rate_id: int,
                     status: Optional[UserRateStatus] = None,
                     score: Optional[int] = None,
                     chapters: Optional[int] = None,
                     episodes: Optional[int] = None,
                     volumes: Optional[int] = None,
                     rewatches: Optional[int] = None,
                     text: Optional[str] = None):
        """Updates user rate and return new info about it.

        :param rate_id: ID of user rate to edit
        :type rate_id: int

        :param status: Status of target
        :type status: Optional[UserRateStatus]

        :param score: Score of target
        :type score: Optional[int]

        :param chapters: Watched/read chapters of target
        :type chapters: Optional[int]

        :param episodes: Watched/read episodes of target
        :type episodes: Optional[int]

        :param volumes: Watched/read volumes of target
        :type volumes: Optional[int]

        :param rewatches: Number of target rewatches
        :type rewatches: Optional[int]

        :param text: Text note for user rate
        :type text: Optional[str]

        :return: Info about new user rate
        :rtype: Optional[UserRate]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           status=status,
                                           score=score,
                                           chapters=chapters,
                                           episodes=episodes,
                                           volumes=volumes,
                                           rewatches=rewatches,
                                           text=text)

        response = await self._client.request(
            self._client.endpoints.user_rate(rate_id),
            data=data_dict,
            request_type=RequestType.PATCH)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=UserRate)

    @method_endpoint('/api/v2/user_rates/:id/increment')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def increment(self, rate_id: int):
        """Increments user rate episode/chapters and return updated info.

        :param rate_id: ID of user rate to increment episode/chapters
        :type rate_id: int

        :return: Info about updated user rate
        :rtype: Optional[UserRate]
        """
        response = await self._client.request(
            self._client.endpoints.user_rate_increment(rate_id),
            request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=UserRate)

    @method_endpoint('/api/v2/user_rates/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete(self, rate_id: int):
        """Deletes user rate.

        :param rate_id: ID of user rate to delete
        :type rate_id: int

        :return: Status of user rate deletion
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.user_rate(rate_id),
            request_type=RequestType.DELETE)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.NO_CONTENT)

    @method_endpoint('/api/user_rates/:type/cleanup')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete_all(self, user_rate_type: UserRateType):
        """Deletes all user rates.

        :param user_rate_type: Type of user rates to delete
        :type user_rate_type: UserRateType

        :return: Status of user rates deletion
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.user_rates_cleanup(str(user_rate_type)),
            request_type=RequestType.DELETE)

        logger.info(response)

        return True

    @method_endpoint('/api/user_rates/:type/reset')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def reset_all(self, user_rate_type: UserRateType):
        """Resets all user rates.

        :param user_rate_type: Type of user rates to reset
        :type user_rate_type: UserRateType

        :return: Status of user rates reset
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.user_rates_reset(str(user_rate_type)),
            request_type=RequestType.DELETE)

        logger.info(response)

        return True



"""Represents `/api/users` and `/api/v2/users` resources."""
from typing import Any, Dict, List, Optional, Union, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import (AnimeCensorship, AnimeList, HistoryTargetType, MessageType,
                     RequestType)
from ..exceptions import ShikimoriAPIResponseError
from ..models import (Ban, ClubInfo, Favourites, History, Message,
                      UnreadMessages, User, UserInfo, UserBrief, UserList)
from ..utils import Utils
from .base_resource import BaseResource


class Users(BaseResource):
    """Users resource class.

    Used to represent `/api/users` and `/api/v2/users` resources
    """

    @method_endpoint('/api/users')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      page: Optional[int] = None,
                      limit: Optional[int] = None):
        """Returns list of users.

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :return: List of users
        :rtype: List[UserInfo]
        """
        query_dict = Utils.create_query_dict(page=page, limit=limit)

        response = await self._client.request(self._client.endpoints.users,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=UserInfo)

    @method_endpoint('/api/users/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, user_id: Union[str, int]):
        """Returns info about user.

        :param user_id: User ID/Nickname to get info
        :type user_id: Union[str, int]

        :return: Info about user
        :rtype: Optional[User]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname)

        response = await self._client.request(
            self._client.endpoints.user(user_id), query=query_dict)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=User)

    @method_endpoint('/api/users/:id/info')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def info(self, user_id: Union[str, int]):
        """Returns user's brief info.

        :param user_id: User ID/Nickname to get brief info
        :type user_id: Union[int, str]

        :return: User's brief info
        :rtype: Optional[UserBrief]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname)

        response = await self._client.request(
            self._client.endpoints.user_info(user_id), query=query_dict)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=UserBrief)

    @method_endpoint('/api/users/whoami')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def current(self):
        """Returns brief info about current user.

        Current user evaluated depending on authorization code.

        :return: Current user brief info
        :rtype: Optional[UserBrief]
        """
        response = await self._client.request(self._client.endpoints.whoami)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=UserBrief)

    @method_endpoint('/api/users/sign_out')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def sign_out(self):
        """Sends sign out request to API.

        :return: True if request was successful, False otherwise
        :rtype: bool
        """
        response = await self._client.request(self._client.endpoints.sign_out)

        return cast(str, response) == 'signed out'

    @method_endpoint('/api/users/:id/friends')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def friends(self, user_id: Union[str, int]):
        """Returns user's friends.

        :param user_id: User ID/Nickname to get friends
        :type user_id: Union[int, str]

        :return: List of user's friends
        :rtype: List[UserInfo]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname)

        response = await self._client.request(
            self._client.endpoints.user_friends(user_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=UserInfo)

    @method_endpoint('/api/users/:id/clubs')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def clubs(self, user_id: Union[int, str]):
        """Returns user's clubs.

        :param user_id: User ID/Nickname to get clubs
        :type user_id: Union[int, str]

        :return: List of user's clubs
        :rtype: List[ClubInfo]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname)

        response = await self._client.request(
            self._client.endpoints.user_clubs(user_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=ClubInfo)

    @method_endpoint('/api/users/:id/anime_rates')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def anime_rates(self,
                          user_id: Union[int, str],
                          page: Optional[int] = None,
                          limit: Optional[int] = None,
                          status: Optional[AnimeList] = None,
                          censored: Optional[AnimeCensorship] = None):
        """Returns user's anime list.

        :param user_id: User ID/Nickname to get anime list
        :type user_id: Optional[int, str]

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param status: Status of status of anime in list
        :type status: Optional[AnimeList]

        :param censored: Type of anime censorship
        :type censored: Optional[AnimeCensorship]

        :return: User's anime list
        :rtype: List[UserList]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname,
                                             page=page,
                                             limit=limit,
                                             status=status,
                                             censored=censored)

        response = await self._client.request(
            self._client.endpoints.user_anime_rates(user_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=UserList)

    @method_endpoint('/api/users/:id/manga_rates')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def manga_rates(self,
                          user_id: Union[int, str],
                          page: Optional[int] = None,
                          limit: Optional[int] = None,
                          censored: Optional[AnimeCensorship] = None):
        """Returns user's manga list.

        :param user_id: User ID/Nickname to get manga list
        :type user_id: Union[int, str]

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param censored: Type of manga censorship
        :type censored: Optional[AnimeCensorship]

        :return: User's manga list
        :rtype: List[UserList]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname,
                                             page=page,
                                             limit=limit,
                                             censored=censored)

        response = await self._client.request(
            self._client.endpoints.user_manga_rates(user_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=UserList)

    @method_endpoint('/api/users/:id/favourites')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def favourites(self, user_id: Union[int, str]):
        """Returns user's favourites.

        :param user_id: User ID/Nickname to get favourites
        :type user_id: Union[int, str]

        :return: User's favourites
        :rtype: Optional[Favourites]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname)

        response = await self._client.request(
            self._client.endpoints.user_favourites(user_id), query=query_dict)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Favourites)

    @method_endpoint('/api/users/:id/messages')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def messages(self,
                       user_id: Union[int, str],
                       page: Optional[int] = None,
                       limit: Optional[int] = None,
                       message_type: MessageType = MessageType.NEWS):
        """Returns current user's messages by type.

        :param user_id: Current user ID/Nickname to get messages
        :type user_id: Union[int, str]

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of page limits
        :type limit: Optional[int]

        :param message_type: Type of message
        :type message_type: MessageType

        :return: Current user's messages
        :rtype: List[Message]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname,
                                             page=page,
                                             limit=limit,
                                             type=message_type)

        response = await self._client.request(
            self._client.endpoints.user_messages(user_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Message)

    @method_endpoint('/api/users/:id/unread_messages')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def unread_messages(
            self, user_id: Union[int, str]) -> Optional[UnreadMessages]:
        """Returns current user's unread messages counter.

        :param user_id: Current user ID/Nickname to get unread messages
        :type user_id: Union[int, str]

        :return: Current user's unread messages counters
        :rtype: Optional[UnreadMessages]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname)

        response = await self._client.request(
            self._client.endpoints.user_unread_messages(user_id),
            query=query_dict)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=UnreadMessages)

    @method_endpoint('/api/users/:id/history')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def history(self,
                      user_id: Union[int, str],
                      page: Optional[int] = None,
                      limit: Optional[int] = None,
                      target_id: Optional[int] = None,
                      target_type: Optional[HistoryTargetType] = None):
        """Returns history of user.

        :param user_id: User ID/Nickname to get history
        :type user_id: Union[int, str]

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param target_id: ID of anime/manga in history
        :type target_id: Optional[int]

        :param target_type: Type of target
        :type target_type: Optional[HistoryTargetType]

        :return: User's history
        :rtype: List[History]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname,
                                             page=page,
                                             limit=limit,
                                             target_id=target_id,
                                             target_type=target_type)

        response = await self._client.request(
            self._client.endpoints.user_history(user_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=History)

    @method_endpoint('/api/users/:id/bans')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def bans(self, user_id: Union[int, str]):
        """Returns list of bans of user.

        :param user_id: User ID/Nickname to get list of bans
        :type user_id: Union[int, str]

        :return: User's bans
        :rtype: List[Ban]
        """
        is_nickname = True if isinstance(user_id, str) else None
        query_dict = Utils.create_query_dict(is_nickname=is_nickname)

        response = await self._client.request(
            self._client.endpoints.user_bans(user_id), query=query_dict)
        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Ban)

    @method_endpoint('/api/v2/users/:user_id/ignore')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def ignore(self, user_id: int):
        """Sets user as ignored.

        :param user_id: ID of topic to ignore
        :type user_id: int

        :return: True if user was ignored, False otherwise
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.user_ignore(user_id),
            request_type=RequestType.POST)

        return cast(Dict[str, Any], response).get('is_ignored') is True

    @method_endpoint('/api/v2/users/:user_id/ignore')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def unignore(self, user_id: int):
        """Sets user as unignored.

        :param user_id: ID of user to unignore
        :type user_id: int

        :return: True if user was unignored, False otherwise
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.user_ignore(user_id),
            request_type=RequestType.DELETE)

        return cast(Dict[str, Any], response).get('is_ignored') is False



"""Model for `/api/users`"""
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from .stats import Stats
from .user_image import UserImage


class UserInfo(BaseModel):
    """Represents user basic info entity."""
    id: int
    nickname: str
    avatar: str
    image: UserImage
    last_online_at: datetime
    url: str


class UserBrief(UserInfo):
    """Represents user brief info entity."""
    name: Optional[str]
    sex: Optional[str]
    full_years: Optional[int]
    locale: Optional[str]


class User(UserBrief):
    """Represents user full info entity."""
    last_online: str
    website: str
    birth_on: Optional[datetime]
    location: Optional[str]
    banned: bool
    about: str
    about_html: str
    common_info: List[str]
    show_comments: bool
    in_friends: Optional[bool]
    is_ignored: bool
    stats: Stats
    style_id: int



"""Submodel for `franchise_tree.py`."""
from pydantic import BaseModel


class TreeLink(BaseModel):
    """Represents tree link entity."""
    id: int
    source_id: int
    target_id: int
    source: int
    target: int
    weight: int
    relation: str



"""Model for `/api/calendar`."""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from .anime import AnimeInfo


class CalendarEvent(BaseModel):
    """Represents event entity in events calendar."""
    next_episode: int
    next_episode_at: datetime
    duration: Optional[int]
    anime: AnimeInfo



"""Submodel for `franchise_tree.py`."""
from typing import Optional

from pydantic import BaseModel


class TreeNode(BaseModel):
    """Represents tree node entity."""
    id: int
    date: int
    name: str
    image_url: str
    url: str
    year: Optional[int]
    kind: Optional[str]
    weight: int



"""Enums for `/api/comments`."""
from .enhanced_enum import EnhancedEnum


class CommentableType(EnhancedEnum):
    """Contains constants related for commentable type
    of fetched comment."""
    TOPIC = 'Topic'
    USER = 'User'


class CommentableCreateType(EnhancedEnum):
    """Contains constants related for commentable type
    of created comment."""
    TOPIC = 'Topic'
    USER = 'User'
    ANIME = 'Anime'
    MANGA = 'Manga'
    CHARACTER = 'Character'
    PERSON = 'Person'
    ARTICLE = 'Article'
    CLUB = 'Club'
    CLUB_PAGE = 'ClubPage'
    COLLECTION = 'Collection'
    CRITIQUE = 'Critique'
    REVIEW = 'Review'



"""Model for `/api/people` and submodel for `creator.py`."""
from datetime import datetime
from typing import List, Optional, Tuple

from pydantic import BaseModel

from .date import Date
from .image import Image
from .roles import Roles
from .works import Works


class PersonInfo(BaseModel):
    """Represents person info entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str


class Person(PersonInfo):
    """Represents person entity."""
    japanese: str
    job_title: str
    birth_on: Date
    deceased_on: Date
    website: str
    groupped_roles: List[Tuple[str, int]]
    roles: List[Roles]
    works: List[Works]
    topic_id: Optional[int]
    person_favoured: bool
    producer: bool
    producer_favoured: bool
    mangaka: bool
    mangaka_favoured: bool
    seyu: bool
    seyu_favoured: bool
    updated_at: datetime
    thread_id: Optional[int]
    # ? Seems like it's gonna be removed soon
    # because of birth_on and deceased_on fields
    birthday: Date



"""Contains package version and some magic for importing API object."""
from .api import ShikimoriAPI
from .store import JSONStore, MemoryStore, NullStore, Store

__version__ = '2.5.1'
__all__ = ['ShikimoriAPI', 'Store', 'NullStore', 'MemoryStore', 'JSONStore']



"""Submodel for `stats.py`."""
from typing import List

from pydantic import BaseModel

from .status import Status


class StatusList(BaseModel):
    """Represents status collection of anime/manga."""
    anime: List[Status]
    manga: List[Status]



"""Memory based config store class"""
from typing import Optional

from ..exceptions import StoreException
from .base import ConfigsDict, ReturnConfig, Store, Token


class MemoryStore(Store):
    """Memory config store class.

    This class is used for storing configs in RAM for faster access
    """

    __slots__ = ('_configs',)

    def __init__(self, configs: Optional[ConfigsDict] = None):
        super().__init__()
        self._configs: ConfigsDict = configs or {}

    @property
    def configs(self) -> ConfigsDict:
        return self._configs

    async def save_config(self,
                          app_name: str,
                          client_id: str,
                          client_secret: str,
                          redirect_uri: str,
                          scopes: str,
                          access_token: str,
                          refresh_token: Optional[str] = None,
                          token_expire_at: Optional[int] = None,
                          auth_code: Optional[str] = None):
        app_data = self._configs.get(app_name)
        if app_data is None:
            app_data = self._configs[app_name] = {}

        app_token: Token = {
            'auth_code': auth_code,
            'scopes': scopes,
            'access_token': access_token,
            'refresh_token': refresh_token,
            'token_expire_at': token_expire_at
        }

        app_tokens = app_data.get('tokens', [])
        for token in app_tokens:
            if token.get('auth_code') == auth_code and token.get(
                    'scopes') == scopes:
                token.update(app_token)
                break
        else:
            app_tokens.append(app_token)

        # Replaced update with keyword arguments
        # as mypy has type check problems with it
        # https://github.com/python/mypy/issues/6019
        app_data.update({
            'client_id': client_id,
            'client_secret': client_secret,
            'redirect_uri': redirect_uri,
            'tokens': app_tokens
        })

    async def fetch_by_access_token(self, app_name: str, access_token: str):
        app_config = self._configs.get(app_name)

        if app_config is None:
            return None

        for token in app_config['tokens']:
            if token.get('access_token') == access_token:
                config: ReturnConfig = {
                    'app_name': app_name,
                    'client_id': app_config['client_id'],
                    'client_secret': app_config['client_secret'],
                    'redirect_uri': app_config['redirect_uri'],
                    # See: https://github.com/python/mypy/issues/9408
                    **token  # type: ignore[misc]
                }
                return config

        return None

    async def fetch_by_auth_code(self, app_name: str, auth_code: str):
        app_config = self._configs.get(app_name)
        if app_config is None:
            return None

        for token in app_config['tokens']:
            if token.get('auth_code') == auth_code:
                config: ReturnConfig = {
                    'app_name': app_name,
                    'client_id': app_config['client_id'],
                    'client_secret': app_config['client_secret'],
                    'redirect_uri': app_config['redirect_uri'],
                    # See: https://github.com/python/mypy/issues/9408
                    **token  # type: ignore[misc]
                }
                return config

        return None

    async def delete_token(self, app_name: str, access_token: str):
        app_config = self._configs.get(app_name)

        if app_config is None:
            raise StoreException(f'The access token {access_token} \
                            of the "{app_name}" app does not exist')

        if len(app_config['tokens']) < 2:
            return await self.delete_all_tokens(app_name)

        for idx, token in enumerate(app_config['tokens']):
            if token.get('access_token') == access_token:
                del app_config['tokens'][idx]

    async def delete_all_tokens(self, app_name: str):
        try:
            self._configs.pop(app_name)
        except KeyError as exc:
            raise StoreException(
                f'The "{app_name}" config app does not exist') from exc

    async def close(self):
        self._configs.clear()
        return await super().close()



"""Base class for shikithon API class."""
from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from time import time
from typing import (Any, AsyncIterator, Awaitable, Dict, List, Optional,
                    TypedDict, TypeVar, Union, cast)

import backoff
from aiohttp import ClientSession, ContentTypeError, FormData
from loguru import logger
from pyrate_limiter import Duration, Limiter, RequestRate

from .endpoints import Endpoints
from .enums import RequestType, ResponseCode
from .exceptions import (AlreadyRunningClient, InvalidContentType,
                         MissingAppVariable, RetryLater,
                         ShikimoriAPIResponseError, ShikithonException)
from .store import NullStore, Store
from .utils import Utils

MAX_CALLS_PER_SECOND = 5
MAX_CALLS_PER_MINUTE = 90

SHIKIMORI_API_URL = 'https://shikimori.me/api'
SHIKIMORI_API_URL_V2 = 'https://shikimori.me/api/v2'
SHIKIMORI_OAUTH_URL = 'https://shikimori.me/oauth'
DEFAULT_REDIRECT_URI = 'urn:ietf:wg:oauth:2.0:oob'

RT = TypeVar('RT')


class ClientConfig(TypedDict, total=False):
    app_name: str
    client_id: str
    client_secret: str
    redirect_uri: str
    scopes: str
    auth_code: Optional[str]
    access_token: str
    refresh_token: Optional[str]
    token_expire_at: Optional[int]


class TokensDict(TypedDict):
    access_token: str
    refresh_token: str
    token_type: str
    scope: str
    created_at: int
    expires_in: int


second_rate = RequestRate(MAX_CALLS_PER_SECOND, Duration.SECOND)
minute_rate = RequestRate(MAX_CALLS_PER_MINUTE, Duration.MINUTE)
request_limiter = Limiter(second_rate, minute_rate)


class Client:
    """Base client class for shikithon API class.

    Contains logic and methods for making requests to the shikimori API
    as well as validating config and etc
    """

    __slots__ = ('endpoints', '_app_name', '_store', '_auto_close_store',
                 '_session', '_config')

    def __init__(self,
                 app_name: str = 'Api Test',
                 store: Store = NullStore(),
                 auto_close_store: bool = True):
        self._app_name = app_name
        self._store = store
        self.endpoints = Endpoints(SHIKIMORI_API_URL, SHIKIMORI_API_URL_V2,
                                   SHIKIMORI_OAUTH_URL)

        self._session: Optional[ClientSession] = None
        self._config: Optional[ClientConfig] = None
        self._auto_close_store = auto_close_store

    @property
    def closed(self):
        """Checks if session is closed.

        :return: True if session is closed, False otherwise
        :rtype: bool
        """
        return self._session is None

    @property
    def restricted_mode(self):
        """Returns current restrict mode of client object.

        If true, client object can access only public methods

        :return: Current restrict mode
        :rtype: bool
        """
        return self._config is None

    @property
    def store(self):
        """Returns store object.

        :return: Store object
        :rtype: Store
        """
        return self._store

    def _set_user_agent(self, app_name: Optional[str]):
        """Updates session headers and set user agent.

        :param app_name: OAuth App name
        :type app_name: Optional[str]
        """
        if self._session is None:
            return

        if app_name is None:
            self._session.headers.pop('User-Agent', None)
            return
        self._session.headers.update({'User-Agent': app_name})

    user_agent = property(fset=_set_user_agent)

    def _set_authorization_header(self, access_token: Optional[str]):
        """Sets authorization header to current session.

        :param access_token: Access token
        :type access_token: Optional[str]
        """
        if self._session is None:
            return

        if access_token is None:
            self._session.headers.pop('Authorization', None)
            return
        self._session.headers.update(
            {'Authorization': 'Bearer ' + access_token})

    authorization_header = property(fset=_set_authorization_header)

    @property
    def config(self) -> Optional[ClientConfig]:
        """Returns current config.

        If config is not availble, returns None

        :return: Current config
        :rtype: Optional[ClientConfig]
        """
        return self._config

    @config.setter
    def config(self, config: Optional[ClientConfig]):
        """Sets new config.

        If passed config isn't valid, raises Exception

        :param config: New config data
        :type config: Optional[ClientConfig]
        """
        if config is None:
            return None

        self.validate_config(config)

        self._config = config

    def validate_config(self, config: ClientConfig):
        """Validates passed config.

        Method checks config for required dict keys

        :param config: Config to validate
        :type config: ClientConfig

        :return: True if config is valid, False otherwise
        :rtype: bool

        :raises MissingAppVariable: If config is invalid and raises is True
        """
        if not config.get('app_name'):
            raise MissingAppVariable('app_name')

        if not config.get('client_id'):
            raise MissingAppVariable('client_id')

        if not config.get('client_secret'):
            raise MissingAppVariable('client_secret')

        if not config.get('auth_code'):
            raise MissingAppVariable('auth_code')

        if not config.get('access_token'):
            raise MissingAppVariable('access_token')

        if not config.get('redirect_uri'):
            config['redirect_uri'] = DEFAULT_REDIRECT_URI

        if not config.get('scopes'):
            config['scopes'] = ''

        return True

    @asynccontextmanager
    async def auth(self,
                   app_name: Optional[str] = None,
                   client_id: Optional[str] = None,
                   client_secret: Optional[str] = None,
                   auth_code: Optional[str] = None,
                   access_token: Optional[str] = None,
                   refresh_token: Optional[str] = None,
                   token_expire_at: Optional[int] = None,
                   redirect_uri: str = DEFAULT_REDIRECT_URI,
                   scopes: str = '') -> AsyncIterator[Client]:
        """Async context manager for authentification.

        :param app_name: OAuth App name
        :type app_name: Optional[str]

        :param client_id: OAuth App client id
        :type client_id: Optional[str]

        :param client_secret: OAuth App client secret
        :type client_secret: Optional[str]

        :param auth_code: OAuth App auth code
        :type auth_code: Optional[str]

        :param access_token: OAuth App access token
        :type access_token: Optional[str]

        :param refresh_token: OAuth App refresh token
        :type refresh_token: Optional[str]

        :param token_expire_at: OAuth App token expire time
        :type token_expire_at: Optional[int]

        :param redirect_uri: OAuth App redirect uri
        :type redirect_uri: str

        :param scopes: OAuth App scopes
        :type scopes: str

        :return: Async iterator of client object
        :rtype: AsyncIterator[Client]

        :raises AlreadyRunningClient: If client is already running
        :raises MissingAppVariable: If some of required variables is missing
        """
        if not self.closed:
            raise AlreadyRunningClient()

        if app_name is None:
            app_name = self._app_name

        if self.store.closed:
            await self.store.open()

        try:
            async with self:
                if access_token is not None:
                    self.config = await self.store.fetch_by_access_token(
                        app_name, access_token)
                elif auth_code is not None:
                    self.config = await self.store.fetch_by_auth_code(
                        app_name, auth_code)

                if self.config is None:
                    if client_id is None or client_secret is None:
                        raise MissingAppVariable(['client_id', 'client_secret'])

                    if access_token is not None:
                        self.config = {
                            'app_name': app_name,
                            'client_id': client_id,
                            'client_secret': client_secret,
                            'redirect_uri': redirect_uri,
                            'auth_code': auth_code,
                            'scopes': scopes,
                            'access_token': access_token,
                            'refresh_token': refresh_token,
                            'token_expire_at': token_expire_at
                        }
                    elif auth_code is not None:
                        token_data = await self.get_access_token(
                            client_id, client_secret, auth_code, redirect_uri)
                        self.config = {
                            'app_name':
                                app_name,
                            'client_id':
                                client_id,
                            'client_secret':
                                client_secret,
                            'redirect_uri':
                                redirect_uri,
                            'auth_code':
                                auth_code,
                            'scopes':
                                token_data['scope'],
                            'access_token':
                                token_data['access_token'],
                            'refresh_token':
                                token_data['refresh_token'],
                            'token_expire_at':
                                token_data['created_at'] +
                                token_data['expires_in']
                        }
                    else:
                        raise MissingAppVariable(['auth_code', 'access_token'])

                    await self.store.save_config(**self.config)

                self.user_agent = self.config['app_name']
                self.authorization_header = self.config['access_token']

                yield self
        finally:
            self._config = None
            if self._auto_close_store and not self.store.closed:
                await self.store.close()

    async def get_access_token(self, client_id: str, client_secret: str,
                               auth_code: str, redirect_uri: str) -> TokensDict:
        """Returns new access token.

        :param client_id: Client ID
        :type client_id: str

        :param client_secret: Client secret
        :type client_secret: str

        :param auth_code: Authorization code
        :type auth_code: str

        :param redirect_uri: Redirect URI
        :type redirect_uri: str

        :return: New access token
        :rtype: TokensDict
        """
        logger.info('Getting new access token')

        data_dict = {
            'grant_type': 'authorization_code',
            'client_id': client_id,
            'client_secret': client_secret,
            'code': auth_code,
            'redirect_uri': redirect_uri
        }

        tokens = await self.request(self.endpoints.oauth_token,
                                    data=data_dict,
                                    request_type=RequestType.POST,
                                    output_logging=False)
        return cast(TokensDict, tokens)

    async def refresh_access_token(
            self,
            client_id: str,
            client_secret: str,
            refresh_token: Optional[str] = None) -> TokensDict:
        """Refreshes expired access token.

        :param client_id: Client ID
        :type client_id: str

        :param client_secret: Client secret
        :type client_secret: str

        :param refresh_token: Refresh token
        :type refresh_token: Optional[str]

        :return: Refreshed access token
        :rtype: TokensDict
        """
        logger.info('Refreshing current access token')

        if refresh_token is None:
            raise ShikithonException('Missing refresh_token. Returning None')

        data_dict = {
            'grant_type': 'refresh_token',
            'client_id': client_id,
            'client_secret': client_secret,
            'refresh_token': refresh_token
        }

        tokens = await self.request(self.endpoints.oauth_token,
                                    data=data_dict,
                                    request_type=RequestType.POST,
                                    output_logging=False)
        return cast(TokensDict, tokens)

    def token_expired(self, token_expire_at: int):
        """Checks if current access token is expired.

        :return: True if token is expired, False otherwise
        :rtype: bool
        """
        logger.debug('Checking if token is expired')

        token_expiration_status = int(time()) > token_expire_at

        logger.debug(f'Token expire status: {token_expiration_status}')
        return token_expiration_status

    @backoff.on_exception(backoff.expo, RetryLater, max_time=10, max_tries=20)
    async def request(
        self,
        url: str,
        data: Optional[Union[Dict[str, Dict[str, str]], Dict[str, str]]] = None,
        form_data: Optional[FormData] = None,
        query: Optional[Dict[str, str]] = None,
        request_type: RequestType = RequestType.GET,
        output_logging: bool = True,
    ) -> Optional[Union[Any, int]]:
        """Creates request and returns response JSON.

        This method uses request_limiter library for rate limiting
        requests (Shikimori API limit: 90rpm and 5rps)

        To address duplication of methods
        for different request methods, this method
        uses RequestType enum

        :param url: URL for making request
        :type url: str

        :param data: Request body data
        :type data: Optional[Union[Dict[str, Dict[str, str]], Dict[str, str]]]

        :param form_data: Form data for multipart/form-data requests
        :type form_data: Optional[FormData]

        :param query: Query data for request
        :type query: Optional[Dict[str, str]]

        :param request_type: Type of current request
        :type request_type: RequestType

        :param output_logging: Parameter for logging JSON response
        :type output_logging: bool

        :return: Response JSON, status code or None
        :rtype: Optional[Union[Any, int]]

        :raises RetryLater: If Shikimori API returns 429 status code
        :raises ShikimoriAPIResponseError: If response status is
        not lower than 400
        :raises InvalidContentType: If response content type not JSON
        """
        if self.closed or self._session is None:
            return None

        logger.info(
            f'{request_type.value} {url}' \
            f'{Utils.convert_to_query_string(query)}')
        if output_logging:
            logger.debug(f'Request info details: {data=}, {query=}')

        if self._is_protected_request(url):
            token_expire_at = None if self.config is None else self.config.get(
                'token_expire_at')
            if isinstance(token_expire_at,
                          int) and self.token_expired(token_expire_at):
                await self._refresh_and_save_tokens()

        async with request_limiter.ratelimit('request', delay=True):
            if request_type == RequestType.GET:
                response = await self._session.get(url, params=query)
            elif request_type == RequestType.POST:
                response = await self._session.post(url,
                                                    data=form_data,
                                                    json=data,
                                                    params=query)
            elif request_type == RequestType.PUT:
                response = await self._session.put(url,
                                                   data=form_data,
                                                   json=data,
                                                   params=query)
            elif request_type == RequestType.PATCH:
                response = await self._session.patch(url,
                                                     data=form_data,
                                                     json=data,
                                                     params=query)
            elif request_type == RequestType.DELETE:
                response = await self._session.delete(url,
                                                      json=data,
                                                      params=query)
            else:
                logger.debug('Unknown request type passed. Returning None')
                return None

        await Utils.log_response_info(response, not output_logging)

        try:
            if response.status == 401 and self._is_protected_request(url):
                await self._refresh_and_save_tokens()
                return await self.request(url, data, form_data, query,
                                          request_type, output_logging)
            elif response.status == ResponseCode.RETRY_LATER.value:
                raise RetryLater('Hit retry later code. Retrying backoff')
            elif not response.ok:
                raise ShikimoriAPIResponseError(
                    method=response.method,
                    status=response.status,
                    url=repr(response.request_info.real_url),
                    text=await response.text())

            logger.debug('Check if response has empty body')
            response_text = await response.text()
            if response_text == '':
                logger.debug('Response has empty body. ' \
                    'Returning response status')
                return response.status
            elif response_text == 'null':
                logger.debug('Response is "null". Returning None')
                return None


            logger.debug('Response is not empty. ' \
                'Trying to extract JSON from response')
            try:
                json_response = await response.json()
            except ContentTypeError:
                # Special case for such method, like
                # /api/users/sign_out
                if response.content_type == 'text/plain':
                    logger.debug('Failed JSON extracting.' \
                        ' Getting response text')
                    return response_text
                logger.error("Response content type isn't valid JSON")
                raise InvalidContentType(response.content_type) from None

            if json_response is None or json_response == {}:
                logger.debug('JSON is empty. ' \
                    'Returning response status')
                return response.status

            logger.debug('Successful extraction. ' \
                    'Returning extracted data')
            return json_response
        finally:
            response.release()

    async def multiple_requests(self,
                                requests: List[Awaitable[RT]]) -> List[RT]:
        """Makes multiple requests to API at the same time.

        :param requests: List with async requests
        :type requests: List[Awaitable[RT]]

        :return: List of responses
        :rtype: List[RT]
        """
        if self.closed:
            return []

        logger.info(f'Gathering {len(requests)} requests')

        return await asyncio.gather(*requests, return_exceptions=False)

    async def _refresh_and_save_tokens(self):
        """Refreshes current access token and saves it to the store.

        Due to some problems when trying to refresh with
        Authorization header, this method sets the header to None
        before refreshing and then sets it back to the new token
        """
        if self._config is None:
            return None

        self.authorization_header = None

        token_data = await self.refresh_access_token(
            self._config['client_id'], self._config['client_secret'],
            self._config['refresh_token'])

        self.authorization_header = token_data['access_token']

        self._config.update({
            'access_token':
                token_data['access_token'],
            'refresh_token':
                token_data['refresh_token'],
            'token_expire_at':
                token_data['created_at'] + token_data['expires_in'],
        })

        self.validate_config(self._config)

        await self.store.save_config(**self._config)

    def _is_protected_request(self, url: str):
        """Checks if a protected request is being made.

        :param url: Request URL
        :type url: str

        :return: True if request is protected, False otherwise
        :rtype: bool
        """
        return not self.restricted_mode and \
                url != self.endpoints.oauth_token and \
                (self.config is not None and \
                    self.config['refresh_token'] is not None)

    async def open(self):
        """Opens session and returns self.

        :return: Client instance
        :rtype: Client
        """
        if self.closed:
            self._session = ClientSession()
            self.user_agent = self._app_name

        return self

    async def close(self):
        """Closes session."""
        if self.closed or self._session is None:
            return

        await self._session.close()
        self._session = None

    async def __aenter__(self):
        """Async context manager entry point.

        :return: Client instance
        :rtype: Client
        """
        return await self.open()

    async def __aexit__(self, *args):
        """Async context manager exit point."""
        await self.close()



"""Represents `/api/animes` and `/api/animes/:anime_id/videos` resources."""
from typing import Any, Dict, List, Optional, Union, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import (AnimeCensorship, AnimeDuration, AnimeKind, AnimeList,
                     AnimeOrder, AnimeRating, AnimeStatus, AnimeTopicKind,
                     RequestType, ResponseCode, VideoKind)
from ..exceptions import ShikimoriAPIResponseError
from ..models import (AnimeInfo, Anime, FranchiseTree, Link, Relation, Role,
                      Screenshot, Topic, Video)
from ..utils import Utils
from .base_resource import BaseResource

DICT_NAME = 'video'


class Animes(BaseResource):
    """Anime resource class.

    Used to represent `/api/animes` and
    `/api/animes/:anime_id/videos` resources
    """

    @method_endpoint('/api/animes')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      page: Optional[int] = None,
                      limit: Optional[int] = None,
                      order: Optional[AnimeOrder] = None,
                      kind: Optional[Union[AnimeKind, List[AnimeKind]]] = None,
                      status: Optional[Union[AnimeStatus,
                                             List[AnimeStatus]]] = None,
                      season: Optional[Union[str, List[str]]] = None,
                      score: Optional[int] = None,
                      duration: Optional[Union[AnimeDuration,
                                               List[AnimeDuration]]] = None,
                      rating: Optional[Union[AnimeRating,
                                             List[AnimeRating]]] = None,
                      genre: Optional[Union[int, List[int]]] = None,
                      studio: Optional[Union[int, List[int]]] = None,
                      franchise: Optional[Union[int, List[int]]] = None,
                      censored: Optional[AnimeCensorship] = None,
                      mylist: Optional[Union[AnimeList,
                                             List[AnimeList]]] = None,
                      ids: Optional[Union[int, List[int]]] = None,
                      exclude_ids: Optional[Union[int, List[int]]] = None,
                      search: Optional[str] = None):
        """Returns list of anime.

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param order: Type of order in list
        :type order: Optional[AnimeOrder]

        :param kind: Type(s) of anime topics
        :type kind: Optional[Union[AnimeKind, List[AnimeKind]]]

        :param status: Type(s) of anime status
        :type status: Optional[Union[AnimeStatus, List[AnimeStatus]]]

        :param season: Name(s) of anime seasons
        :type season: Optional[Union[str, List[str]]]

        :param score: Minimal anime score to filter
        :type score: Optional[int]

        :param duration: Duration size(s) of anime
        :type duration: Optional[Union[AnimeDuration, List[AnimeDuration]]]

        :param rating: Type of anime rating(s)
        :type rating: Optional[Union[AnimeRating, List[AnimeRating]]]

        :param genre: Genre(s) ID
        :type genre: Optional[Union[int, List[int]]]

        :param studio: Studio(s) ID
        :type studio: Optional[Union[int, List[int]]]

        :param franchise: Franchise(s) ID
        :type franchise: Optional[Union[int, List[int]]]

        :param censored: Type of anime censorship
        :type censored: Optional[AnimeCensorship]

        :param mylist: Status(-es) of anime in current user list.
            If app is in restricted mode,
            this parameter won't affect on response.
        :type mylist: Optional[Union[AnimeList, List[AnimeList]]]

        :param ids: Anime(s) ID to include
        :type ids: Optional[Union[int, List[int]]]

        :param exclude_ids: Anime(s) ID to exclude
        :type exclude_ids: Optional[Union[int, List[int]]]

        :param search: Search phrase to filter animes by name
        :type search: Optional[str]

        :return: List of anime
        :rtype: List[AnimeInfo]
        """
        query_dict = Utils.create_query_dict(page=page,
                                             limit=limit,
                                             order=order,
                                             kind=kind,
                                             status=status,
                                             season=season,
                                             score=score,
                                             duration=duration,
                                             rating=rating,
                                             genre=genre,
                                             studio=studio,
                                             franchise=franchise,
                                             censored=censored,
                                             mylist=mylist,
                                             ids=ids,
                                             exclude_ids=exclude_ids,
                                             search=search)

        response = await self._client.request(self._client.endpoints.animes,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=AnimeInfo)

    @method_endpoint('/api/animes/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, anime_id: int):
        """Returns info about certain anime.

        :param anime_id: Anime ID to get info
        :type anime_id: int

        :return: Anime info
        :rtype: Optional[Anime]
        """
        response = await self._client.request(
            self._client.endpoints.anime(anime_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Anime)

    @method_endpoint('/api/animes/:id/roles')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def roles(self, anime_id: int):
        """Returns roles of certain anime.

        :param anime_id: Anime ID to get roles
        :type anime_id: int

        :return: List of anime roles
        :rtype: List[Role]
        """
        response = await self._client.request(
            self._client.endpoints.anime_roles(anime_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Role)

    @method_endpoint('/api/animes/:id/similar')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def similar(self, anime_id: int):
        """Returns list of similar animes for certain anime.

        :param anime_id: Anime ID to get similar animes
        :type anime_id: int

        :return: List of similar animes
        :rtype: List[AnimeInfo]
        """
        response = await self._client.request(
            self._client.endpoints.similar_animes(anime_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=AnimeInfo)

    @method_endpoint('/api/animes/:id/related')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def related(self, anime_id: int):
        """Returns list of relations of certain anime.

        :param anime_id: Anime ID to get relations
        :type anime_id: int

        :return: List of relations
        :rtype: List[Relation]
        """
        response = await self._client.request(
            self._client.endpoints.anime_related_content(anime_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Relation)

    @method_endpoint('/api/animes/:id/screenshots')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def screenshots(self, anime_id: int):
        """Returns list of screenshot links of certain anime.

        :param anime_id: Anime ID to get screenshot links
        :type anime_id: int

        :return: List of screenshot links
        :rtype: List[Screenshot]
        """
        response = await self._client.request(
            self._client.endpoints.anime_screenshots(anime_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Screenshot)

    @method_endpoint('/api/animes/:id/franchise')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def franchise(self, anime_id: int):
        """Returns franchise tree of certain anime.

        :param anime_id: Anime ID to get franchise tree
        :type anime_id: int

        :return: Franchise tree of certain anime
        :rtype: Optional[FranchiseTree]
        """
        response = await self._client.request(
            self._client.endpoints.anime_franchise_tree(anime_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=FranchiseTree)

    @method_endpoint('/api/animes/:id/external_links')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def external_links(self, anime_id: int):
        """Returns list of external links of certain anime.

        :param anime_id: Anime ID to get external links
        :type anime_id: int

        :return: List of external links
        :rtype: List[Link]
        """
        response = await self._client.request(
            self._client.endpoints.anime_external_links(anime_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Link)

    @method_endpoint('/api/animes/:id/topics')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def topics(self,
                     anime_id: int,
                     page: Optional[int] = None,
                     limit: Optional[int] = None,
                     kind: Optional[AnimeTopicKind] = None,
                     episode: Optional[int] = None):
        """Returns anime's list of topics.

        :param anime_id: Anime ID to get topics
        :type anime_id: int

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param kind: Kind of topic
        :type kind: Optional[AnimeTopicKind]

        :param episode: Number of anime episode
        :type episode: Optional[int]

        :return: List of topics
        :rtype: List[Topic]
        """
        query_dict = Utils.create_query_dict(page=page,
                                             limit=limit,
                                             kind=kind,
                                             episode=episode)

        response = await self._client.request(
            self._client.endpoints.anime_topics(anime_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Topic)

    @method_endpoint('/api/animes/:anime_id/videos')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def videos(self, anime_id: int):
        """Returns list of anime videos.

        :param anime_id: Anime ID to get videos
        :type anime_id: int

        :return: Anime videos list
        :rtype: List[Video]
        """
        response = await self._client.request(
            self._client.endpoints.anime_videos(anime_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Video)

    @method_endpoint('/api/animes/:anime_id/videos')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def create_video(self, anime_id: int, kind: VideoKind, name: str,
                           url: str):
        """Creates anime video.

        :param anime_id: Anime ID to create video
        :type anime_id: int

        :param kind: Kind of video
        :type kind: VideoKind

        :param name: Name of video
        :type name: str

        :param url: URL of video
        :type url: str

        :return: Created video info
        :rtype: Optional[Video]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           kind=kind,
                                           name=name,
                                           url=url)

        response = await self._client.request(
            self._client.endpoints.anime_videos(anime_id),
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Video)

    @method_endpoint('/api/animes/:anime_id/videos/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete_video(self, anime_id: int, video_id: int):
        """Deletes anime video.

        :param anime_id: Anime ID to delete video
        :type anime_id: int

        :param video_id: Video ID
        :type video_id: str

        :return: Status of video deletion
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.anime_video(anime_id, video_id),
            request_type=RequestType.DELETE)

        return Utils.validate_response_code(cast(int, response),
                                            ResponseCode.SUCCESS)



"""Submodel for `creator.py`."""
from datetime import datetime
from typing import List, Optional, Union

from pydantic import BaseModel

from .anime import CharacterAnime
from .image import Image
from .manga import CharacterManga
from .ranobe import CharacterRanobe
from .seyu import Seyu


class CharacterInfo(BaseModel):
    """Represents character info entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str


class Character(CharacterInfo):
    """Represents character entity."""
    altname: str
    japanese: str
    description: Optional[str]
    description_html: str
    description_source: Optional[str]
    favoured: bool
    thread_id: Optional[int]
    topic_id: Optional[int]
    updated_at: datetime
    seyu: List[Seyu]
    animes: List[CharacterAnime]
    mangas: List[Union[CharacterManga, CharacterRanobe]]



"""Exception for raising on invalid request response content type."""

from loguru import logger


class InvalidContentType(Exception):

    def __init__(self, content_type: str):
        error_message = f'Invalid response content type: {content_type}'
        logger.error(error_message)
        super().__init__(error_message)



"""Represents `/api/reviews` resource."""
from typing import Any, Dict, Optional, cast

from loguru import logger

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RequestType, ReviewOpinion
from ..exceptions import ShikimoriAPIResponseError
from ..models import Review
from ..utils import Utils
from .base_resource import BaseResource

DICT_NAME = 'review'


class Reviews(BaseResource):
    """Reviews resource class.

    Used to represent `/api/reviews` resource
    """

    @method_endpoint('/api/reviews')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def create_anime(self, anime_id: int, body: str,
                           opinion: ReviewOpinion):
        """Creates review for anime.

        :param anime_id: ID of anime to create review for
        :type anime_id: int

        :param body: Body of review
        :type body: str

        :param opinion: Opinion of the review
        :type opinion: ReviewOpinion

        :return: Created review info
        :rtype: Optional[Review]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           anime_id=anime_id,
                                           body=body,
                                           opinion=opinion)

        response = await self._client.request(self._client.endpoints.reviews,
                                              data=data_dict,
                                              request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Review)

    # @method_endpoint('/api/reviews')
    # @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    # async def create_manga(self,
    #                        manga_id: int,
    #                        body: str,
    #                        opinion: ReviewOpinion
    # ):
    #     """Creates review for manga."""
    # ! Currently there is no way to use API for creating review for manga
    # ? Unfortunately, trying to send request with "manga_id"
    # ? and without "anime_id", returns error, that "anime_id" is missing
    # * So this method is comment out for later addition, when API will be
    # * capable of creating review for manga, tho source code has lines for that
    # * See reviews_controller.rb in Shikimori repository

    @method_endpoint('/api/reviews/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def update(self,
                     review_id: int,
                     body: Optional[str] = None,
                     opinion: Optional[ReviewOpinion] = None,
                     is_written_before_release: Optional[bool] = None):
        """Updates review.

        :param review_id: ID of review to update
        :type review_id: int

        :param body: New body of review
        :type body: Optional[str]

        :param opinion: New opinion of review
        :type opinion: Optional[ReviewOpinion]

        :param is_written_before_release:
        Status of review publish, before title release
        :type is_written_before_release: Optional[bool]

        :return: Updated review info
        or None if review cannot be updated
        :rtype: Optional[Review]
        """
        data_dict = Utils.create_data_dict(
            dict_name=DICT_NAME,
            body=body,
            opinion=opinion,
            is_written_before_release=is_written_before_release)

        response = await self._client.request(
            self._client.endpoints.review(review_id),
            data=data_dict,
            request_type=RequestType.PATCH)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Review)

    @method_endpoint('/api/reviews/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete(self, review_id: int):
        """Deletes review.

        :param review_id: ID of review to delete
        :type review_id: int

        :return: Status of review deletion
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.review(review_id),
            request_type=RequestType.DELETE)

        logger.info(response)

        return True



"""Submodel for `user.py`."""
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field

from .activity import Activity
from .genre import Genre
from .publisher import Publisher
from .rating_list import RatingList
from .score_list import ScoreList
from .status_list import StatusList
from .studio import Studio
from .type_list import TypeList


class Stats(BaseModel):
    """Represents user's stats entity."""
    statuses: Optional[StatusList]
    full_statuses: Optional[StatusList]
    scores: Optional[ScoreList]
    types: Optional[TypeList]
    ratings: Optional[RatingList]
    has_anime: Optional[bool] = Field(alias='has_anime?')
    has_manga: Optional[bool] = Field(alias='has_manga?')
    genres: List[Genre]
    studios: List[Studio]
    publishers: List[Publisher]
    activity: Optional[Union[List[Activity], Dict]]



"""Represents `/api/publishers` resource."""
from typing import Any, Dict, List, cast

from ..decorators import exceptions_handler, method_endpoint
from ..exceptions import ShikimoriAPIResponseError
from ..models import Publisher
from ..utils import Utils
from .base_resource import BaseResource


class Publishers(BaseResource):
    """Publishers resource class.

    Used to represent `/api/publishers` resource
    """

    @method_endpoint('/api/publishers')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self):
        """Returns list of publishers.

        :return: List of publishers
        :rtype: List[Publisher]
        """
        response = await self._client.request(self._client.endpoints.publishers)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Publisher)



"""Exceptions for shikithon API class.

Used for raising on validation of config/app data, request errors, etc.
"""

from .already_running_client import AlreadyRunningClient
from .invalid_content_type import InvalidContentType
from .missing_app_variable import MissingAppVariable
from .retry_later import RetryLater
from .shikimori_api_response_error import ShikimoriAPIResponseError
from .shikithon_exception import ShikithonException
from .store_exception import StoreException

__all__ = [
    'AlreadyRunningClient', 'MissingAppVariable', 'RetryLater',
    'StoreException', 'ShikimoriAPIResponseError', 'InvalidContentType',
    'ShikithonException'
]
